@@ -6,38 +6,39 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#include <asm/types.h>          /* For __uXX types */
+#include <asm/types.h>		/* For __uXX types */
 
-#define IP_VS_VERSION_CODE            0x010009
-#define NVERSION(version)                       \
-	(version >> 16) & 0xFF,                 \
-	(version >> 8) & 0xFF,                  \
+#define IP_VS_VERSION_CODE	0x010100
+#define NVERSION(version)			\
+	(version >> 16) & 0xFF,			\
+	(version >> 8) & 0xFF,			\
 	version & 0xFF
 
 /*
  *      Virtual Service Flags
  */
-#define IP_VS_SVC_F_PERSISTENT        0x0001    /* persistent port */
-#define IP_VS_SVC_F_HASHED            0x0002    /* hashed entry */
+#define IP_VS_SVC_F_PERSISTENT	0x0001		/* persistent port */
+#define IP_VS_SVC_F_HASHED	0x0002		/* hashed entry */
 
 /*
  *      Destination Server Flags
  */
-#define IP_VS_DEST_F_AVAILABLE        0x0001    /* Available tag */
+#define IP_VS_DEST_F_AVAILABLE	0x0001		/* server is available */
+#define IP_VS_DEST_F_OVERLOAD	0x0002		/* server is overloaded */
 
 /*
  *      IPVS sync daemon states
  */
-#define IP_VS_STATE_NONE    0           /* daemon is stopped */
-#define IP_VS_STATE_MASTER  1           /* started as master */
-#define IP_VS_STATE_BACKUP  2           /* started as backup */
+#define IP_VS_STATE_NONE	0		/* daemon is stopped */
+#define IP_VS_STATE_MASTER	1		/* started as master */
+#define IP_VS_STATE_BACKUP	2		/* started as backup */
 
 /*
  *      IPVS socket options
  */
 #define IP_VS_BASE_CTL		(64+1024+64)		/* base */
 
-#define IP_VS_SO_SET_NONE	IP_VS_BASE_CTL	        /* just peek */
+#define IP_VS_SO_SET_NONE	IP_VS_BASE_CTL		/* just peek */
 #define IP_VS_SO_SET_INSERT	(IP_VS_BASE_CTL+1)
 #define IP_VS_SO_SET_ADD	(IP_VS_BASE_CTL+2)
 #define IP_VS_SO_SET_EDIT	(IP_VS_BASE_CTL+3)
@@ -47,7 +48,7 @@
 #define IP_VS_SO_SET_ADDDEST	(IP_VS_BASE_CTL+7)
 #define IP_VS_SO_SET_DELDEST	(IP_VS_BASE_CTL+8)
 #define IP_VS_SO_SET_EDITDEST	(IP_VS_BASE_CTL+9)
-#define IP_VS_SO_SET_TIMEOUTS	(IP_VS_BASE_CTL+10)
+#define IP_VS_SO_SET_TIMEOUT	(IP_VS_BASE_CTL+10)
 #define IP_VS_SO_SET_STARTDAEMON (IP_VS_BASE_CTL+11)
 #define IP_VS_SO_SET_STOPDAEMON (IP_VS_BASE_CTL+12)
 #define IP_VS_SO_SET_RESTORE    (IP_VS_BASE_CTL+13)
@@ -61,7 +62,7 @@
 #define IP_VS_SO_GET_SERVICE	(IP_VS_BASE_CTL+3)
 #define IP_VS_SO_GET_DESTS	(IP_VS_BASE_CTL+4)
 #define IP_VS_SO_GET_DEST	(IP_VS_BASE_CTL+5)	/* not used now */
-#define IP_VS_SO_GET_TIMEOUTS	(IP_VS_BASE_CTL+6)
+#define IP_VS_SO_GET_TIMEOUT	(IP_VS_BASE_CTL+6)
 #define IP_VS_SO_GET_DAEMON	(IP_VS_BASE_CTL+7)
 #define IP_VS_SO_GET_MAX	IP_VS_SO_GET_DAEMON
 
@@ -69,50 +70,55 @@
 /*
  *      IPVS Connection Flags
  */
-#define IP_VS_CONN_F_FWD_MASK         0x0007    /* mask for the fwd methods */
-#define IP_VS_CONN_F_MASQ	      0x0000    /* masquerading */
-#define IP_VS_CONN_F_LOCALNODE	      0x0001    /* local node */
-#define IP_VS_CONN_F_TUNNEL	      0x0002    /* tunneling */
-#define IP_VS_CONN_F_DROUTE           0x0003    /* direct routing */
-#define IP_VS_CONN_F_BYPASS           0x0004    /* cache bypass */
-#define IP_VS_CONN_F_HASHED	      0x0040	/* hashed entry */
-#define IP_VS_CONN_F_NOOUTPUT         0x0080    /* no output packets */
-#define IP_VS_CONN_F_INACTIVE         0x0100    /* not established */
-#define IP_VS_CONN_F_OUT_SEQ          0x0200    /* must do output seq adjust */
-#define IP_VS_CONN_F_IN_SEQ           0x0400    /* must do input seq adjust */
-#define IP_VS_CONN_F_SEQ_MASK         0x0600    /* in/out sequence mask */
-#define IP_VS_CONN_F_NO_CPORT         0x0800    /* no client port set yet */
+#define IP_VS_CONN_F_FWD_MASK	0x0007		/* mask for the fwd methods */
+#define IP_VS_CONN_F_MASQ	0x0000		/* masquerading/NAT */
+#define IP_VS_CONN_F_LOCALNODE	0x0001		/* local node */
+#define IP_VS_CONN_F_TUNNEL	0x0002		/* tunneling */
+#define IP_VS_CONN_F_DROUTE	0x0003		/* direct routing */
+#define IP_VS_CONN_F_BYPASS	0x0004		/* cache bypass */
+#define IP_VS_CONN_F_SYNC	0x0020		/* entry created by sync */
+#define IP_VS_CONN_F_HASHED	0x0040		/* hashed entry */
+#define IP_VS_CONN_F_NOOUTPUT	0x0080		/* no output packets */
+#define IP_VS_CONN_F_INACTIVE	0x0100		/* not established */
+#define IP_VS_CONN_F_OUT_SEQ	0x0200		/* must do output seq adjust */
+#define IP_VS_CONN_F_IN_SEQ	0x0400		/* must do input seq adjust */
+#define IP_VS_CONN_F_SEQ_MASK	0x0600		/* in/out sequence mask */
+#define IP_VS_CONN_F_NO_CPORT	0x0800		/* no client port set yet */
 
 /* Move it to better place one day, for now keep it unique */
 #define NFC_IPVS_PROPERTY	0x10000
 
-#define IP_VS_SCHEDNAME_MAXLEN         16
-#define IP_VS_IFNAME_MAXLEN            16
+#define IP_VS_SCHEDNAME_MAXLEN	16
+#define IP_VS_IFNAME_MAXLEN	16
 
-struct ip_vs_rule_user {
-	/* global options */
-	int             tcp_timeout;    /* timeout values */
-	int             tcp_fin_timeout;
-	int             udp_timeout;
-	int             state;          /* sync daemon state */
-	char            mcast_ifn[IP_VS_IFNAME_MAXLEN];
-					/* multicast interface name */
+
+struct ip_vs_service_user {
+	/* virtual service addresses */
+	u_int16_t		protocol;
+	u_int32_t		addr;		/* virtual ip address */
+	u_int16_t		port;
+	u_int32_t		fwmark;		/* firwall mark of service */
 
 	/* virtual service options */
-	u_int16_t	protocol;
-	u_int32_t	vaddr;          /* virtual address */
-	u_int16_t	vport;
-	u_int32_t       vfwmark;        /* firwall mark of virtual service*/
-	char            sched_name[IP_VS_SCHEDNAME_MAXLEN];
-	unsigned	vs_flags;       /* virtual service flags */
-	unsigned        timeout;        /* persistent timeout in ticks */
-	u_int32_t	netmask;        /* persistent netmask */
+	char			sched_name[IP_VS_SCHEDNAME_MAXLEN];
+	unsigned		flags;		/* virtual service flags */
+	unsigned		timeout;	/* persistent timeout in sec */
+	u_int32_t		netmask;	/* persistent netmask */
+};
 
-	/* destination specific options */
-	u_int32_t	daddr;          /* destination address */
-	u_int16_t	dport;
-	unsigned        conn_flags;     /* destination flags */
-	int             weight;         /* destination weight */
+
+struct ip_vs_dest_user {
+	/* destination server address */
+	u_int32_t		addr;
+	u_int16_t		port;
+
+	/* real server options */
+	unsigned		conn_flags;	/* connection flags */
+	int			weight;		/* destination weight */
+
+	/* thresholds for active connections */
+	u_int32_t		u_threshold;	/* upper threshold */
+	u_int32_t		l_threshold;	/* lower threshold */
 };
 
 
@@ -138,83 +144,95 @@ struct ip_vs_stats_user
 /* The argument to IP_VS_SO_GET_INFO */
 struct ip_vs_getinfo {
 	/* version number */
-	unsigned int	version;
+	unsigned int		version;
 
 	/* size of connection hash table */
-	unsigned int	size;
+	unsigned int		size;
 
 	/* number of virtual services */
-	unsigned int	num_services;
+	unsigned int		num_services;
 };
 
+
 /* The argument to IP_VS_SO_GET_SERVICE */
-struct ip_vs_service_user {
-	/* which service: user fills this in */
-	u_int16_t	protocol;
-	u_int32_t	addr;           /* virtual address */
-	u_int16_t	port;
-	u_int32_t       fwmark;         /* firwall mark of virtual service */
+struct ip_vs_service_entry {
+	/* which service: user fills in these */
+	u_int16_t		protocol;
+	u_int32_t		addr;		/* virtual address */
+	u_int16_t		port;
+	u_int32_t		fwmark;		/* firwall mark of service */
 
 	/* service options */
-	char            sched_name[IP_VS_SCHEDNAME_MAXLEN];
-	unsigned	flags;          /* virtual service flags */
-	unsigned        timeout;        /* persistent timeout in ticks */
-	u_int32_t	netmask;        /* persistent netmask */
+	char			sched_name[IP_VS_SCHEDNAME_MAXLEN];
+	unsigned		flags;          /* virtual service flags */
+	unsigned		timeout;	/* persistent timeout */
+	u_int32_t		netmask;	/* persistent netmask */
 
 	/* number of real servers */
-	unsigned int    num_dests;
+	unsigned int		num_dests;
 
 	/* statistics */
 	struct ip_vs_stats_user stats;
 };
 
-struct ip_vs_dest_user {
-	u_int32_t	addr;           /* destination address */
-	u_int16_t	port;
-	unsigned	flags;		/* destination flags */
-	int		weight;         /* destination weight */
-	u_int32_t	activeconns;	/* active connections */
-	u_int32_t	inactconns;	/* inactive connections */
+
+struct ip_vs_dest_entry {
+	u_int32_t		addr;		/* destination address */
+	u_int16_t		port;
+	unsigned		conn_flags;	/* connection flags */
+	int			weight;		/* destination weight */
+	u_int32_t		activeconns;	/* active connections */
+	u_int32_t		inactconns;	/* inactive connections */
+
+	u_int32_t		u_threshold;	/* upper threshold */
+	u_int32_t		l_threshold;	/* lower threshold */
 
 	/* statistics */
 	struct ip_vs_stats_user stats;
 };
 
+
 /* The argument to IP_VS_SO_GET_DESTS */
 struct ip_vs_get_dests {
-	/* which service: user fills this in */
-	u_int16_t	protocol;
-	u_int32_t	addr;           /* virtual address */
-	u_int16_t	port;
-	u_int32_t       fwmark;         /* firwall mark of virtual service */
+	/* which service: user fills in these */
+	u_int16_t		protocol;
+	u_int32_t		addr;		/* virtual address */
+	u_int16_t		port;
+	u_int32_t		fwmark;		/* firwall mark of service */
 
 	/* number of real servers */
-	unsigned int    num_dests;
+	unsigned int		num_dests;
 
 	/* the real servers */
-	struct ip_vs_dest_user entrytable[0];
+	struct ip_vs_dest_entry	entrytable[0];
 };
 
+
 /* The argument to IP_VS_SO_GET_SERVICES */
 struct ip_vs_get_services {
 	/* number of virtual services */
-	unsigned int num_services;
+	unsigned int		num_services;
 
 	/* service table */
-	struct ip_vs_service_user entrytable[0];
+	struct ip_vs_service_entry entrytable[0];
 };
 
-/* The argument to IP_VS_SO_GET_TIMEOUTS */
+
+/* The argument to IP_VS_SO_GET_TIMEOUT */
 struct ip_vs_timeout_user {
-	int             tcp_timeout;
-	int             tcp_fin_timeout;
-	int             udp_timeout;
+	int			tcp_timeout;
+	int			tcp_fin_timeout;
+	int			udp_timeout;
 };
 
+
 /* The argument to IP_VS_SO_GET_DAEMON */
 struct ip_vs_daemon_user {
-	int	state;				/* sync daemon state */
-	char	mcast_ifn[IP_VS_IFNAME_MAXLEN];	/* multicast interface name */
+	/* sync daemon state (master/backup) */
+	int			state;
+
+	/* multicast interface name */
+	char			mcast_ifn[IP_VS_IFNAME_MAXLEN];
 };
 
 
@@ -226,11 +244,11 @@ struct ip_vs_daemon_user {
 #include <linux/skbuff.h>               /* for struct sk_buff */
 #include <linux/ip.h>                   /* for struct iphdr */
 #include <asm/atomic.h>                 /* for struct atomic_t */
-#include <linux/netdevice.h>		/* for struct neighbour; */
+#include <linux/netdevice.h>		/* for struct neighbour */
 #include <net/dst.h>			/* for struct dst_entry */
-#include <net/route.h>			/* for ip_route_output */
 #include <net/tcp.h>
 #include <net/udp.h>
+#include <linux/compiler.h>
 
 
 #ifdef CONFIG_IP_VS_DEBUG
@@ -245,9 +263,22 @@ extern int ip_vs_get_debug_level(void);
 	    if (net_ratelimit())			\
 		    printk(KERN_DEBUG "IPVS: " msg);	\
     } while (0)
+#define IP_VS_DBG_PKT(level, pp, iph, msg)		\
+    do {						\
+	    if (level <= ip_vs_get_debug_level())	\
+		pp->debug_packet(pp, iph, msg);		\
+    } while (0)
+#define IP_VS_DBG_RL_PKT(level, pp, iph, msg)		\
+    do {						\
+	    if (level <= ip_vs_get_debug_level() &&	\
+		net_ratelimit())			\
+		pp->debug_packet(pp, iph, msg);		\
+    } while (0)
 #else	/* NO DEBUGGING at ALL */
 #define IP_VS_DBG(level, msg...)  do {} while (0)
 #define IP_VS_DBG_RL(msg...)  do {} while (0)
+#define IP_VS_DBG_PKT(level, pp, iph, msg)	do {} while (0)
+#define IP_VS_DBG_RL_PKT(level, pp, iph, msg)	do {} while (0)
 #endif
 
 #define IP_VS_BUG() BUG()
@@ -279,6 +310,8 @@ extern int ip_vs_get_debug_level(void);
 #define LeaveFunction(level)   do {} while (0)
 #endif
 
+#define	IP_VS_WAIT_WHILE(expr)	while (expr) { cpu_relax(); }
+
 
 /*
  *      The port number of FTP service (in network order).
@@ -286,7 +319,6 @@ extern int ip_vs_get_debug_level(void);
 #define FTPPORT  __constant_htons(21)
 #define FTPDATA  __constant_htons(20)
 
-
 /*
  *      IPVS sysctl variables under the /proc/sys/net/ipv4/vs/
  */
@@ -320,44 +352,45 @@ enum {
 	NET_IPV4_VS_LAST
 };
 
-
 /*
- *      IPVS State Values
+ *      TCP State Values
  */
 enum {
-	IP_VS_S_NONE = 0,
-	IP_VS_S_ESTABLISHED,
-	IP_VS_S_SYN_SENT,
-	IP_VS_S_SYN_RECV,
-	IP_VS_S_FIN_WAIT,
-	IP_VS_S_TIME_WAIT,
-	IP_VS_S_CLOSE,
-	IP_VS_S_CLOSE_WAIT,
-	IP_VS_S_LAST_ACK,
-	IP_VS_S_LISTEN,
-	IP_VS_S_SYNACK,
-	IP_VS_S_UDP,
-	IP_VS_S_ICMP,
-	IP_VS_S_LAST
+	IP_VS_TCP_S_NONE = 0,
+	IP_VS_TCP_S_ESTABLISHED,
+	IP_VS_TCP_S_SYN_SENT,
+	IP_VS_TCP_S_SYN_RECV,
+	IP_VS_TCP_S_FIN_WAIT,
+	IP_VS_TCP_S_TIME_WAIT,
+	IP_VS_TCP_S_CLOSE,
+	IP_VS_TCP_S_CLOSE_WAIT,
+	IP_VS_TCP_S_LAST_ACK,
+	IP_VS_TCP_S_LISTEN,
+	IP_VS_TCP_S_SYNACK,
+	IP_VS_TCP_S_LAST
 };
 
-
-struct ip_vs_timeout_table {
-	atomic_t refcnt;
-	int scale;
-	int timeout[IP_VS_S_LAST+1];
+/* UDP State Values */
+enum {
+	IP_VS_UDP_S_NORMAL,
+	IP_VS_UDP_S_LAST,
 };
 
+/* ICMP State Values */
+enum {
+	IP_VS_ICMP_S_NORMAL,
+	IP_VS_ICMP_S_LAST,
+};
 
 /*
  *	Transport protocol header
  */
 union ip_vs_tphdr {
-	unsigned char *raw;
-	struct udphdr *uh;
-	struct tcphdr *th;
-	struct icmphdr *icmph;
-	__u16 *portp;
+	unsigned char		*raw;
+	struct udphdr		*uh;
+	struct tcphdr		*th;
+	struct icmphdr		*icmph;
+	__u16			*portp;
 };
 
 
@@ -367,10 +400,10 @@ union ip_vs_tphdr {
  *      Only used in the VS/NAT.
  */
 struct ip_vs_seq {
-	__u32           init_seq;       /* Add delta from this seq */
-	__u32           delta;          /* Delta in sequence numbers */
-	__u32           previous_delta; /* Delta in sequence numbers
-					   before last resized pkt */
+	__u32			init_seq;	/* Add delta from this seq */
+	__u32			delta;		/* Delta in sequence numbers */
+	__u32			previous_delta;	/* Delta in sequence numbers
+						   before last resized pkt */
 };
 
 
@@ -394,6 +427,78 @@ struct ip_vs_stats
 	spinlock_t              lock;           /* spin lock */
 };
 
+struct ip_vs_conn;
+struct ip_vs_app;
+
+#define	CONFIG_IP_VS_PROTO_UDP
+#define	CONFIG_IP_VS_PROTO_TCP
+#define	CONFIG_IP_VS_PROTO_ICMP
+#define	CONFIG_IP_VS_PROTO_ESP
+#define	CONFIG_IP_VS_PROTO_AH
+
+struct ip_vs_protocol {
+	struct ip_vs_protocol	*next;
+	char			*name;
+	__u16			protocol;
+	int			minhlen;
+	int			minhlen_icmp;
+	int			dont_defrag;
+	int			skip_nonexisting;
+	int			slave;		/* if controlled by others */
+	atomic_t		appcnt;		/* counter of proto app incs */
+	int			*timeout_table;	/* protocol timeout table */
+
+	void (*init)(struct ip_vs_protocol *pp);
+
+	void (*exit)(struct ip_vs_protocol *pp);
+
+	int (*conn_schedule)(struct sk_buff *skb, struct ip_vs_protocol *pp,
+			     struct iphdr *iph, union ip_vs_tphdr h,
+			     int *verdict, struct ip_vs_conn **cpp);
+
+	struct ip_vs_conn *
+	(*conn_in_get)(struct sk_buff *skb,
+		       struct ip_vs_protocol *pp, struct iphdr *iph,
+		       union ip_vs_tphdr h, int inverse);
+
+	struct ip_vs_conn *
+	(*conn_out_get)(struct sk_buff *skb,
+			struct ip_vs_protocol *pp, struct iphdr *iph,
+			union ip_vs_tphdr h, int inverse);
+
+	int (*snat_handler)(struct sk_buff *skb,
+			    struct ip_vs_protocol *pp, struct ip_vs_conn *cp,
+			    struct iphdr *iph, union ip_vs_tphdr h, int size);
+
+	int (*dnat_handler)(struct sk_buff *skb,
+			    struct ip_vs_protocol *pp, struct ip_vs_conn *cp,
+			    struct iphdr *iph, union ip_vs_tphdr h, int size);
+
+	int (*csum_check)(struct sk_buff *skb,
+			  struct ip_vs_protocol *pp, struct iphdr *iph,
+			  union ip_vs_tphdr h, int size);
+
+	const char *(*state_name)(int state);
+
+	int (*state_transition)(struct ip_vs_conn *cp, int direction,
+				struct iphdr *iph, union ip_vs_tphdr h,
+				struct ip_vs_protocol *pp);
+
+	int (*register_app)(struct ip_vs_app *inc);
+
+	void (*unregister_app)(struct ip_vs_app *inc);
+
+	int (*app_conn_bind)(struct ip_vs_conn *cp);
+
+	void (*debug_packet)(struct ip_vs_protocol *pp, struct iphdr *iph,
+			     char *msg);
+
+	void (*timeout_change)(struct ip_vs_protocol *pp, int flags);
+
+	int (*set_state_timeout)(struct ip_vs_protocol *pp, char *sname, int to);
+};
+
+extern struct ip_vs_protocol * ip_vs_proto_get(unsigned short proto);
 
 /*
  *	IP_VS structure allocated for each dynamically scheduled connection
@@ -414,7 +519,6 @@ struct ip_vs_conn {
 	atomic_t		refcnt;		/* reference count */
 	struct timer_list	timer;		/* Expiration timer */
 	volatile unsigned long	timeout;	/* timeout */
-	struct ip_vs_timeout_table *timeout_table;
 
 	/* Flags and state transition */
 	spinlock_t              lock;           /* lock for state transition */
@@ -428,7 +532,8 @@ struct ip_vs_conn {
 	atomic_t                in_pkts;        /* incoming packet counter */
 
 	/* packet transmitter for different forwarding methods */
-	int (*packet_xmit)(struct sk_buff *skb, struct ip_vs_conn *cp);
+	int (*packet_xmit)(struct sk_buff *skb, struct ip_vs_conn *cp,
+			   struct ip_vs_protocol *pp);
 
 	/* Note: we can group the following members into a structure,
 	   in order to save more space, and the following members are
@@ -461,6 +566,7 @@ struct ip_vs_service {
 	struct list_head	destinations;  /* real server d-linked list */
 	__u32			num_dests;     /* number of servers */
 	struct ip_vs_stats      stats;         /* statistics for the service */
+	struct ip_vs_app	*inc;	  /* bind conns to this app inc */
 
 	/* for scheduling */
 	struct ip_vs_scheduler	*scheduler;    /* bound scheduler object */
@@ -477,27 +583,32 @@ struct ip_vs_dest {
 	struct list_head	n_list;   /* for the dests in the service */
 	struct list_head	d_list;   /* for table with all the dests */
 
-	__u32			addr;	  /* IP address of real server */
-	__u16			port;	  /* port number of the service */
-	unsigned		flags;	  /* dest status flags */
-	atomic_t		weight;	  /* server weight */
+	__u32			addr;		/* IP address of the server */
+	__u16			port;		/* port number of the server */
+	volatile unsigned	flags;		/* dest status flags */
 	atomic_t		conn_flags;	/* flags to copy to conn */
-	atomic_t		activeconns;	/* active connections */
-	atomic_t		inactconns;     /* inactive connections */
-	atomic_t		refcnt;	        /* reference counter */
+	atomic_t		weight;		/* server weight */
+
+	atomic_t		refcnt;		/* reference counter */
 	struct ip_vs_stats      stats;          /* statistics */
 
+	/* connection counters and thresholds */
+	atomic_t		activeconns;	/* active connections */
+	atomic_t		inactconns;	/* inactive connections */
+	__u32			u_threshold;	/* upper threshold */
+	__u32			l_threshold;	/* lower threshold */
+
 	/* for destination cache */
-	spinlock_t		dst_lock;	/* lock dst_cache */
+	spinlock_t		dst_lock;	/* lock of dst_cache */
 	struct dst_entry	*dst_cache;	/* destination cache entry */
 	u32			dst_rtos;	/* RT_TOS(tos) for dst */
 
 	/* for virtual service */
-	struct ip_vs_service    *svc;     /* service that it belongs to */
-	__u16			protocol; /* which protocol (TCP/UDP) */
-	__u32			vaddr;	  /* IP address for virtual service */
-	__u16			vport;	  /* port number for the service */
-	__u32                   vfwmark;  /* firewall mark of the service */
+	struct ip_vs_service	*svc;		/* service it belongs to */
+	__u16			protocol;	/* which protocol (TCP/UDP) */
+	__u32			vaddr;		/* virtual IP address */
+	__u16			vport;		/* virtual port number */
+	__u32			vfwmark;	/* firewall mark of service */
 };
 
 
@@ -505,10 +616,10 @@ struct ip_vs_dest {
  *	The scheduler object
  */
 struct ip_vs_scheduler {
-	struct list_head        n_list;   /* d-linked list head */
-	char			*name;    /* scheduler name */
-	atomic_t                refcnt;   /* reference counter */
-	struct module		*module;  /* THIS_MODULE/NULL */
+	struct list_head	n_list;		/* d-linked list head */
+	char			*name;		/* scheduler name */
+	atomic_t		refcnt;		/* reference counter */
+	struct module		*module;	/* THIS_MODULE/NULL */
 
 	/* scheduler initializing service */
 	int (*init_service)(struct ip_vs_service *svc);
@@ -524,26 +635,65 @@ struct ip_vs_scheduler {
 
 
 /*
- *	The application module object
+ *	The application module object (a.k.a. app incarnation)
  */
 struct ip_vs_app
 {
-	struct list_head        n_list;   /* d-linked list head */
-	char                    *name;    /* name of application module */
-	unsigned                type;     /* type = proto<<16 | port
-					     (host byte order)*/
-	struct module		*module;  /* THIS_MODULE/NULL */
+	struct list_head	a_list;		/* member in app list */
+	int			type;		/* IP_VS_APP_TYPE_xxx */
+	char			*name;		/* application module name */
+	__u16			protocol;
+	struct module		*module;	/* THIS_MODULE/NULL */
+	struct list_head	incs_list;	/* list of incarnations */
+
+	/* members for application incarnations */
+	struct list_head	p_list;		/* member in proto app list */
+	struct ip_vs_app	*app;		/* its real application */
+	__u16			port;		/* port number in net order */
+	atomic_t		usecnt;		/* usage counter */
+
+	/* output hook */
+	int (*pkt_out)(struct ip_vs_app *, struct ip_vs_conn *,
+		       struct sk_buff *);
+
+	/* input hook */
+	int (*pkt_in)(struct ip_vs_app *, struct ip_vs_conn *,
+		      struct sk_buff *);
 
 	/* ip_vs_app initializer */
 	int (*init_conn)(struct ip_vs_app *, struct ip_vs_conn *);
+
 	/* ip_vs_app finish */
 	int (*done_conn)(struct ip_vs_app *, struct ip_vs_conn *);
-	/* output hook */
-	int (*pkt_out)(struct ip_vs_app *,
-		       struct ip_vs_conn *, struct sk_buff *);
-	/* input hook */
-	int (*pkt_in)(struct ip_vs_app *,
-		      struct ip_vs_conn *, struct sk_buff *);
+
+
+	/* not used now */
+	int (*bind_conn)(struct ip_vs_app *, struct ip_vs_conn *,
+			 struct ip_vs_protocol *);
+
+	void (*unbind_conn)(struct ip_vs_app *, struct ip_vs_conn *);
+
+	int *			timeout_table;
+	int *			timeouts;
+	int			timeouts_size;
+
+	int (*conn_schedule)(struct sk_buff *skb, struct ip_vs_app *app,
+			     struct iphdr *iph, union ip_vs_tphdr h,
+			     int *verdict, struct ip_vs_conn **cpp);
+
+	struct ip_vs_conn *
+	(*conn_in_get)(struct sk_buff *skb, struct ip_vs_app *app,
+		       struct iphdr *iph, union ip_vs_tphdr h, int inverse);
+
+	struct ip_vs_conn *
+	(*conn_out_get)(struct sk_buff *skb, struct ip_vs_app *app,
+			struct iphdr *iph, union ip_vs_tphdr h, int inverse);
+
+	int (*state_transition)(struct ip_vs_conn *cp, int direction,
+				struct iphdr *iph,
+				union ip_vs_tphdr h, struct ip_vs_app *app);
+
+	void (*timeout_change)(struct ip_vs_app *app, int flags);
 };
 
 
@@ -554,7 +704,11 @@ struct ip_vs_app
 extern const char *ip_vs_proto_name(unsigned proto);
 extern unsigned int check_for_ip_vs_out(struct sk_buff **skb_p,
 					int (*okfn)(struct sk_buff *));
+extern void ip_vs_init_hash_table(struct list_head *table, int rows);
+#define IP_VS_INIT_HASH_TABLE(t) ip_vs_init_hash_table(t, sizeof(t)/sizeof(t[0]))
 
+#define IP_VS_APP_TYPE_UNSPEC	0
+#define IP_VS_APP_TYPE_FTP	1
 
 /*
  *     ip_vs_conn handling functions
@@ -580,12 +734,22 @@ extern unsigned int check_for_ip_vs_out(struct sk_buff **skb_p,
 #define IP_VS_CONN_TAB_SIZE     (1 << IP_VS_CONN_TAB_BITS)
 #define IP_VS_CONN_TAB_MASK     (IP_VS_CONN_TAB_SIZE - 1)
 
-#define VS_STATE_INPUT	        0
-#define VS_STATE_OUTPUT	        4
-#define VS_STATE_INPUT_ONLY	8
+enum {
+	IP_VS_DIR_INPUT = 0,
+	IP_VS_DIR_OUTPUT,
+	IP_VS_DIR_INPUT_ONLY,
+	IP_VS_DIR_LAST,
+};
 
-extern struct ip_vs_timeout_table vs_timeout_table;
-extern struct ip_vs_timeout_table vs_timeout_table_dos;
+/*
+ *      Slow timer functions for IPVS
+ *      (from ip_vs_timer.c)
+ */
+extern void add_sltimer(struct timer_list * timer);
+extern int  del_sltimer(struct timer_list * timer);
+extern void mod_sltimer(struct timer_list *timer, unsigned long expires);
+extern void ip_vs_sltimer_init(void);
+extern void ip_vs_sltimer_cleanup(void);
 
 extern struct ip_vs_conn *ip_vs_conn_in_get
 (int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port);
@@ -598,6 +762,7 @@ static inline void __ip_vs_conn_put(struct ip_vs_conn *cp)
 	atomic_dec(&cp->refcnt);
 }
 extern void ip_vs_conn_put(struct ip_vs_conn *cp);
+extern void ip_vs_conn_fill_cport(struct ip_vs_conn *cp, __u16 cport);
 
 extern struct ip_vs_conn *
 ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
@@ -605,10 +770,9 @@ ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
 	       struct ip_vs_dest *dest);
 extern void ip_vs_conn_expire_now(struct ip_vs_conn *cp);
 
-extern const char * ip_vs_state_name(int state);
-extern int ip_vs_set_state(struct ip_vs_conn *cp, int state_off,
-			   struct iphdr *iph, void *tp);
-extern int ip_vs_conn_listen(struct ip_vs_conn *cp);
+extern const char * ip_vs_state_name(__u16 proto, int state);
+
+extern void ip_vs_tcp_conn_listen(struct ip_vs_conn *cp);
 extern int ip_vs_check_template(struct ip_vs_conn *ct);
 extern void ip_vs_secure_tcp_set(int on);
 extern void ip_vs_random_dropentry(void);
@@ -668,11 +832,15 @@ ip_vs_control_add(struct ip_vs_conn *cp, struct ip_vs_conn *ctl_cp)
  *      (from ip_vs_app.c)
  */
 #define IP_VS_APP_MAX_PORTS  8
-extern int register_ip_vs_app(struct ip_vs_app *mapp,
-			      unsigned short proto, __u16 port);
-extern int unregister_ip_vs_app(struct ip_vs_app *mapp);
-extern struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp);
-extern int ip_vs_unbind_app(struct ip_vs_conn *cp);
+extern int register_ip_vs_app(struct ip_vs_app *app);
+extern void unregister_ip_vs_app(struct ip_vs_app *app);
+extern int ip_vs_bind_app(struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern void ip_vs_unbind_app(struct ip_vs_conn *cp);
+extern int
+register_ip_vs_app_inc(struct ip_vs_app *app, __u16 proto, __u16 port);
+extern int ip_vs_app_inc_get(struct ip_vs_app *inc);
+extern void ip_vs_app_inc_put(struct ip_vs_app *inc);
+
 extern int ip_vs_app_pkt_out(struct ip_vs_conn *, struct sk_buff *skb);
 extern int ip_vs_app_pkt_in(struct ip_vs_conn *, struct sk_buff *skb);
 extern int ip_vs_skb_replace(struct sk_buff *skb, int pri,
@@ -681,6 +849,22 @@ extern int ip_vs_app_init(void);
 extern void ip_vs_app_cleanup(void);
 
 
+/*
+ *	IPVS protocol functions (from ip_vs_proto.c)
+ */
+extern int ip_vs_protocol_init(void);
+extern void ip_vs_protocol_cleanup(void);
+extern void ip_vs_protocol_timeout_change(int flags);
+extern int *ip_vs_create_timeout_table(int *table, int size);
+extern int
+ip_vs_set_state_timeout(int *table, int num, char **names, char *name, int to);
+extern struct ip_vs_protocol ip_vs_protocol_tcp;
+extern struct ip_vs_protocol ip_vs_protocol_udp;
+extern struct ip_vs_protocol ip_vs_protocol_icmp;
+extern struct ip_vs_protocol ip_vs_protocol_esp;
+extern struct ip_vs_protocol ip_vs_protocol_ah;
+
+
 /*
  *      Registering/unregistering scheduler functions
  *      (from ip_vs_sched.c)
@@ -692,11 +876,14 @@ extern int ip_vs_bind_scheduler(struct ip_vs_service *svc,
 extern int ip_vs_unbind_scheduler(struct ip_vs_service *svc);
 extern struct ip_vs_scheduler *ip_vs_scheduler_get(const char *sched_name);
 extern void ip_vs_scheduler_put(struct ip_vs_scheduler *scheduler);
+extern struct ip_vs_conn *
+ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph);
+extern int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb,
+			struct ip_vs_protocol *pp, union ip_vs_tphdr h);
 
 
 /*
- *      IPVS control data and functions
- *      (from ip_vs_ctl.c)
+ *      IPVS control data and functions (from ip_vs_ctl.c)
  */
 extern int sysctl_ip_vs_cache_bypass;
 extern int sysctl_ip_vs_expire_nodest_conn;
@@ -740,6 +927,24 @@ extern void ip_vs_kill_estimator(struct ip_vs_stats *stats);
 extern void ip_vs_zero_estimator(struct ip_vs_stats *stats);
 
 
+/*
+ *	Various IPVS packet transmitters (from ip_vs_xmit.c)
+ */
+extern int ip_vs_null_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern int ip_vs_bypass_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern int ip_vs_nat_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern int ip_vs_tunnel_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern int ip_vs_dr_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern int ip_vs_icmp_xmit
+(struct sk_buff *skb, struct ip_vs_conn *cp, struct ip_vs_protocol *pp);
+extern void ip_vs_dst_reset(struct ip_vs_dest *dest);
+
+
 /*
  *	This is a simple mechanism to ignore packets when
  *	we are loaded. Just set ip_vs_drop_rate to 'n' and
@@ -756,18 +961,6 @@ static __inline__ int ip_vs_todrop(void)
 	return 1;
 }
 
-
-/*
- *      Slow timer functions for IPVS
- *      (from ip_vs_timer.c)
- */
-extern void add_sltimer(struct timer_list * timer);
-extern int  del_sltimer(struct timer_list * timer);
-extern void mod_sltimer(struct timer_list *timer, unsigned long expires);
-extern void ip_vs_sltimer_init(void);
-extern void ip_vs_sltimer_cleanup(void);
-
-
 /*
  *      ip_vs_fwd_tag returns the forwarding tag of the connection
  */
@@ -795,107 +988,6 @@ extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 }
 
 
-/*
- *	transport layer header checking
- */
-extern inline int ip_vs_header_check(struct sk_buff *skb, int proto, int ihl)
-{
-	int len;
-
-	switch (proto) {
-	case IPPROTO_TCP:
-		len = ihl + sizeof(struct tcphdr);
-		/* we don't care about TCP options */
-		break;
-	case IPPROTO_UDP:
-		len = ihl + sizeof(struct udphdr);
-		break;
-	default:
-		len = 0;
-	}
-
-	/* guarantee protocol header available in skb data area */
-	if (!pskb_may_pull(skb, len))
-		return -1;
-	else
-		return 0;
-}
-
-
-/*
- *      Destination cache
- */
-static inline void
-__ip_vs_dst_set(struct ip_vs_dest *dest, u32 rtos, struct dst_entry *dst)
-{
-	struct dst_entry *old_dst;
-
-	old_dst = dest->dst_cache;
-	dest->dst_cache = dst;
-	dest->dst_rtos = rtos;
-	dst_release(old_dst);
-}
-
-static inline void
-__ip_vs_dst_reset(struct ip_vs_dest *dest)
-{
-	struct dst_entry *old_dst;
-
-	old_dst = dest->dst_cache;
-	dest->dst_cache = NULL;
-	dst_release(old_dst);
-}
-
-static inline struct dst_entry *
-__ip_vs_dst_check(struct ip_vs_dest *dest, u32 rtos, u32 cookie)
-{
-	struct dst_entry *dst = dest->dst_cache;
-
-	if (!dst)
-		return NULL;
-	if ((dst->obsolete || rtos != dest->dst_rtos) &&
-	    dst->ops->check(dst, cookie) == NULL) {
-		dest->dst_cache = 0;
-		return NULL;
-	}
-	dst_hold(dst);
-	return dst;
-}
-
-static inline struct rtable *
-__ip_vs_get_out_rt(struct ip_vs_conn *cp, u32 rtos)
-{
-	struct rtable *rt;			/* Route to the other host */
-	struct ip_vs_dest *dest = cp->dest;
-
-	if (dest) {
-		spin_lock(&dest->dst_lock);
-		if (!(rt = (struct rtable *)
-		      __ip_vs_dst_check(dest, rtos, 0))) {
-			if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
-				spin_unlock(&dest->dst_lock);
-				IP_VS_DBG_RL("ip_route_output error, "
-					     "dest: %u.%u.%u.%u\n",
-					     NIPQUAD(dest->addr));
-				return NULL;
-			}
-			__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
-			IP_VS_DBG(10, "new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
-				  NIPQUAD(dest->addr),
-				  atomic_read(&rt->u.dst.__refcnt), rtos);
-		}
-		spin_unlock(&dest->dst_lock);
-	} else {
-		if (ip_route_output(&rt, cp->daddr, 0, rtos, 0)) {
-			IP_VS_DBG_RL("ip_route_output error, dest: "
-				     "%u.%u.%u.%u\n", NIPQUAD(cp->daddr));
-			return NULL;
-		}
-	}
-
-	return rt;
-}
-
 static inline u16 ip_vs_check_diff(u32 old, u32 new, u16 oldsum)
 {
 	u32 diff[2] = { old, new };
@@ -904,41 +996,6 @@ static inline u16 ip_vs_check_diff(u32 old, u32 new, u16 oldsum)
 				      oldsum ^ 0xFFFF));
 }
 
-static inline void ip_vs_fast_check_update(union ip_vs_tphdr *h,
-	u32 oldip, u32 newip, u16 oldport, u16 newport, u8 protocol)
-{
-	u16 *checkp;
-
-	if (protocol == IPPROTO_TCP)
-		checkp = &h->th->check;
-	else
-		checkp = &h->uh->check;
-	*checkp = ip_vs_check_diff(~oldip, newip,
-		ip_vs_check_diff(oldport ^ 0xFFFF, newport, *checkp));
-	if (!*checkp && protocol == IPPROTO_UDP)
-		*checkp = 0xFFFF;
-}
-
-static inline int
-ip_vs_skb_cow(struct sk_buff *skb, unsigned int headroom,
-	      struct iphdr **iph_p, unsigned char **t_p)
-{
-	int delta = (headroom > 16 ? headroom : 16) - skb_headroom(skb);
-
-	if (delta < 0)
-		delta = 0;
-
-	if (delta ||skb_cloned(skb)) {
-		if (pskb_expand_head(skb, (delta+15)&~15, 0, GFP_ATOMIC))
-			return -ENOMEM;
-
-		/* skb data changed, update pointers */
-		*iph_p = skb->nh.iph;
-		*t_p = (char*) (*iph_p) + (*iph_p)->ihl * 4;
-	}
-	return 0;
-}
-
 #endif /* __KERNEL__ */
 
 #endif	/* _IP_VS_H */
