@@ -1,7 +1,7 @@
 /*
- * IPVS         Application module
+ * ip_vs_app.c: Application module support for IPVS
  *
- * Version:     $Id: ip_vs_app.c,v 1.13 2001/11/04 08:58:43 wensong Exp $
+ * Version:     $Id: ip_vs_app.c,v 1.15 2002/09/15 08:14:08 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -12,7 +12,7 @@
  *
  * Most code here is taken from ip_masq_app.c in kernel 2.2. The difference
  * is that ip_vs_app module handles the reverse direction (incoming requests
- * and outgoing responses). The ip_vs_app modules are only used for VS/NAT.
+ * and outgoing responses).
  *
  *		IP_MASQ_APP application masquerading module
  *
@@ -39,170 +39,253 @@
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
-#define IP_VS_APP_TAB_SIZE  16          /* must be power of 2 */
-
-#define IP_VS_APP_HASH(proto, port) ((port^proto) & (IP_VS_APP_TAB_SIZE-1))
-#define IP_VS_APP_TYPE(proto, port) (proto<<16 | port)
-#define IP_VS_APP_PORT(type)        (type & 0xffff)
-#define IP_VS_APP_PROTO(type)       ((type>>16) & 0x00ff)
-
-
 EXPORT_SYMBOL(register_ip_vs_app);
 EXPORT_SYMBOL(unregister_ip_vs_app);
+EXPORT_SYMBOL(register_ip_vs_app_inc);
+
+/* ipvs application list head */
+static LIST_HEAD(ip_vs_app_list);
+static DECLARE_MUTEX(__ip_vs_app_mutex);
 
 
 /*
- *	will hold ipvs app. hashed list heads
+ *	Get an ip_vs_app object
  */
-static struct list_head ip_vs_app_base[IP_VS_APP_TAB_SIZE];
+static inline int ip_vs_app_get(struct ip_vs_app *app)
+{
+	/* test and MOD_INC_USE_COUNT atomically */
+	if (app->module)
+		return try_inc_mod_count(app->module);
+	else
+		return 1;
+}
 
-/* lock for ip_vs_app table */
-static rwlock_t __ip_vs_app_lock = RW_LOCK_UNLOCKED;
+
+static inline void ip_vs_app_put(struct ip_vs_app *app)
+{
+	if (app->module)
+		__MOD_DEC_USE_COUNT(app->module);
+}
+
+
+/*
+ *	Allocate/initialize app incarnation and register it in proto apps.
+ */
+int
+ip_vs_app_inc_new(struct ip_vs_app *app, __u16 proto, __u16 port)
+{
+	struct ip_vs_protocol *pp;
+	struct ip_vs_app *inc;
+	int ret;
+
+	if (!(pp = ip_vs_proto_get(proto)))
+		return -EPROTONOSUPPORT;
+
+	if (!pp->unregister_app)
+		return -EOPNOTSUPP;
+
+	inc = kmalloc(sizeof(struct ip_vs_app), GFP_KERNEL);
+	if (!inc)
+		return -ENOMEM;
+	memcpy(inc, app, sizeof(*inc));
+	INIT_LIST_HEAD(&inc->p_list);
+	INIT_LIST_HEAD(&inc->incs_list);
+	inc->app = app;
+	inc->port = htons(port);
+	atomic_set(&inc->usecnt, 0);
+
+	if (app->timeouts) {
+		inc->timeout_table =
+			ip_vs_create_timeout_table(app->timeouts,
+						   app->timeouts_size);
+		if (!inc->timeout_table) {
+			ret = -ENOMEM;
+			goto out;
+		}
+	}
+
+	ret = pp->register_app(inc);
+	if (ret)
+		goto out;
+
+	list_add(&inc->a_list, &app->incs_list);
+	IP_VS_DBG(9, "%s application %s:%u registered\n",
+		  pp->name, inc->name, inc->port);
+
+	return 0;
+
+  out:
+	if (inc->timeout_table)
+		kfree(inc->timeout_table);
+	kfree(inc);
+	return ret;
+}
+
+
+/*
+ *	Release app incarnation
+ */
+static void
+ip_vs_app_inc_release(struct ip_vs_app *inc)
+{
+	struct ip_vs_protocol *pp;
+
+	if (!(pp = ip_vs_proto_get(inc->protocol)))
+		return;
+
+	if (pp->unregister_app)
+		pp->unregister_app(inc);
+
+	IP_VS_DBG(9, "%s App %s:%u unregistered\n",
+		  pp->name, inc->name, inc->port);
+
+	list_del(&inc->a_list);
+
+	if (inc->timeout_table != NULL)
+		kfree(inc->timeout_table);
+	kfree(inc);
+}
+
+
+/*
+ *	Get reference to app inc (only called from softirq)
+ *
+ */
+int ip_vs_app_inc_get(struct ip_vs_app *inc)
+{
+	int result;
+
+	atomic_inc(&inc->usecnt);
+	if (unlikely((result = ip_vs_app_get(inc->app)) != 1))
+		atomic_dec(&inc->usecnt);
+	return result;
+}
+
+
+/*
+ *	Put the app inc (only called from timer or net softirq)
+ */
+void ip_vs_app_inc_put(struct ip_vs_app *inc)
+{
+	ip_vs_app_put(inc->app);
+	atomic_inc(&inc->usecnt);
+}
+
+
+/*
+ *	Register an application incarnation in protocol applications
+ */
+int
+register_ip_vs_app_inc(struct ip_vs_app *app, __u16 proto, __u16 port)
+{
+	int result;
+
+	down(&__ip_vs_app_mutex);
+
+	result = ip_vs_app_inc_new(app, proto, port);
+
+	up(&__ip_vs_app_mutex);
+
+	return result;
+}
 
 
 /*
  *	ip_vs_app registration routine
- *	port: host byte order.
  */
-int register_ip_vs_app(struct ip_vs_app *vapp,
-		       unsigned short proto, __u16 port)
+int register_ip_vs_app(struct ip_vs_app *app)
 {
-	unsigned hash;
-
-	if (!vapp) {
-		IP_VS_ERR("register_ip_vs_app(): NULL arg\n");
-		return -EINVAL;
-	}
-
 	MOD_INC_USE_COUNT;
 
-	vapp->type = IP_VS_APP_TYPE(proto, port);
-	hash = IP_VS_APP_HASH(proto, port);
+	down(&__ip_vs_app_mutex);
 
-	write_lock_bh(&__ip_vs_app_lock);
-	list_add(&vapp->n_list, &ip_vs_app_base[hash]);
-	write_unlock_bh(&__ip_vs_app_lock);
+	list_add(&app->a_list, &ip_vs_app_list);
+
+	up(&__ip_vs_app_mutex);
 
 	return 0;
 }
 
 
 /*
- *	ip_vs_app unregistration routine.
+ *	ip_vs_app unregistration routine
+ *	We are sure there are no app incarnations attached to services
  */
-int unregister_ip_vs_app(struct ip_vs_app *vapp)
+void unregister_ip_vs_app(struct ip_vs_app *app)
 {
-	if (!vapp) {
-		IP_VS_ERR("unregister_ip_vs_app(): NULL arg\n");
-		return -EINVAL;
+	struct ip_vs_app *inc;
+	struct list_head *l = &app->incs_list;
+
+	down(&__ip_vs_app_mutex);
+
+	while (l->next != l) {
+		inc = list_entry(l->next, struct ip_vs_app, a_list);
+		ip_vs_app_inc_release(inc);
 	}
 
-	write_lock_bh(&__ip_vs_app_lock);
-	list_del(&vapp->n_list);
-	write_unlock_bh(&__ip_vs_app_lock);
+	list_del(&app->a_list);
+
+	up(&__ip_vs_app_mutex);
 
 	MOD_DEC_USE_COUNT;
-
-	return 0;
 }
 
 
+#if 0000
 /*
- *	get ip_vs_app object by its proto and port (net byte order).
+ *	Get reference to app by name (called from user context)
  */
-static struct ip_vs_app * ip_vs_app_get(unsigned short proto, __u16 port)
+struct ip_vs_app *ip_vs_app_get_by_name(char *appname)
 {
-	struct list_head *e;
-	struct ip_vs_app *vapp;
-	unsigned hash;
-	unsigned type;
+	struct list_head *p;
+	struct ip_vs_app *app, *a = NULL;
 
-	port = ntohs(port);
-	type = IP_VS_APP_TYPE(proto, port);
-	hash = IP_VS_APP_HASH(proto, port);
+	down(&__ip_vs_app_mutex);
 
-	read_lock_bh(&__ip_vs_app_lock);
-
-	list_for_each(e, &ip_vs_app_base[hash]) {
-		vapp = list_entry(e, struct ip_vs_app, n_list);
-
-		/*
-		 * Test and MOD_INC_USE_COUNT atomically
-		 */
-		if (vapp->module && !try_inc_mod_count(vapp->module)) {
-			/*
-			 * This application module is just deleted
-			 */
+	list_for_each (p, &ip_vs_app_list) {
+		app = list_entry(p, struct ip_vs_app, a_list);
+		if (strcmp(app->name, appname))
 			continue;
-		}
-		if (type == vapp->type) {
-			read_unlock_bh(&__ip_vs_app_lock);
-			return vapp;
-		}
 
-		if (vapp->module)
-			__MOD_DEC_USE_COUNT(vapp->module);
+		/* softirq may call ip_vs_app_get too, so the caller
+		   must disable softirq on the current CPU */
+		if (ip_vs_app_get(app))
+			a = app;
+		break;
 	}
 
-	read_unlock_bh(&__ip_vs_app_lock);
-	return NULL;
+	up(&__ip_vs_app_mutex);
+
+	return a;
+}
+#endif
+
+
+/*
+ *	Bind ip_vs_conn to its ip_vs_app (called by cp constructor)
+ */
+int ip_vs_bind_app(struct ip_vs_conn *cp, struct ip_vs_protocol *pp)
+{
+	return pp->app_conn_bind(cp);
 }
 
 
 /*
- *	Bind ip_vs_conn to its ip_vs_app based on proto and dport,
- *	and call the ip_vs_app constructor.
+ *	Unbind cp from application incarnation (called by cp destructor)
  */
-struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp)
+void ip_vs_unbind_app(struct ip_vs_conn *cp)
 {
-	struct ip_vs_app *vapp;
+	struct ip_vs_app *inc = cp->app;
 
-	/* no need to bind app if its forwarding method is not NAT */
-	if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ)
-		return NULL;
+	if (!inc)
+		return;
 
-	if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
-		return NULL;
-
-	/*
-	 *	don't allow binding if already bound
-	 */
-	if (cp->app != NULL) {
-		IP_VS_ERR("ip_vs_bind_app(): "
-			  "called for already bound object.\n");
-		return cp->app;
-	}
-
-	vapp = ip_vs_app_get(cp->protocol, cp->vport);
-
-	if (vapp != NULL) {
-		cp->app = vapp;
-
-		if (vapp->init_conn)
-			vapp->init_conn(vapp, cp);
-	}
-	return vapp;
-}
-
-
-/*
- *	Unbind cp from type object and call cp destructor (does not kfree()).
- */
-int ip_vs_unbind_app(struct ip_vs_conn *cp)
-{
-	struct ip_vs_app *vapp = cp->app;
-
-	if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
-		return 0;
-
-	if (vapp != NULL) {
-		if (vapp->done_conn)
-			vapp->done_conn(vapp, cp);
-		cp->app = NULL;
-		if (vapp->module)
-			__MOD_DEC_USE_COUNT(vapp->module);
-	}
-	return (vapp != NULL);
+	if (inc->unbind_conn)
+		inc->unbind_conn(inc, cp);
+	if (inc->done_conn)
+		inc->done_conn(inc, cp);
+	ip_vs_app_inc_put(inc);
+	cp->app = NULL;
 }
 
 
@@ -272,7 +355,7 @@ static inline void vs_seq_update(struct ip_vs_conn *cp, struct ip_vs_seq *vseq,
 {
 	/* spinlock is to keep updating cp->flags atomic */
 	spin_lock(&cp->lock);
-	if ( !(cp->flags & flag) || after(seq, vseq->init_seq)) {
+	if (!(cp->flags & flag) || after(seq, vseq->init_seq)) {
 		vseq->previous_delta = vseq->delta;
 		vseq->delta += diff;
 		vseq->init_seq = seq;
@@ -284,12 +367,12 @@ static inline void vs_seq_update(struct ip_vs_conn *cp, struct ip_vs_seq *vseq,
 
 /*
  *	Output pkt hook. Will call bound ip_vs_app specific function
- *	called by ip_vs_out(), assumes previously checked cp!=NULL
+ *	called by ipvs packet handler, assumes previously checked cp!=NULL
  *	returns (new - old) skb->len diff.
  */
-int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
+int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff *skb)
 {
-	struct ip_vs_app * vapp;
+	struct ip_vs_app *app;
 	int diff;
 	struct iphdr *iph;
 	struct tcphdr *th;
@@ -299,10 +382,10 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	 *	check if application module is bound to
 	 *	this ip_vs_conn.
 	 */
-	if ((vapp = cp->app) == NULL)
+	if ((app = cp->app) == NULL)
 		return 0;
 
-	iph = (*skb_p)->nh.iph;
+	iph = skb->nh.iph;
 	th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
 
 	/*
@@ -323,10 +406,10 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	/*
 	 *	Call private output hook function
 	 */
-	if (vapp->pkt_out == NULL)
+	if (app->pkt_out == NULL)
 		return 0;
 
-	diff = vapp->pkt_out(vapp, cp, skb_p);
+	diff = app->pkt_out(app, cp, skb);
 
 	/*
 	 *	Update ip_vs seq stuff if len has changed.
@@ -341,12 +424,12 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 
 /*
  *	Input pkt hook. Will call bound ip_vs_app specific function
- *	called by ip_fw_demasquerade(), assumes previously checked cp!=NULL.
+ *	called by ipvs packet handler, assumes previously checked cp!=NULL.
  *	returns (new - old) skb->len diff.
  */
-int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
+int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff *skb)
 {
-	struct ip_vs_app * vapp;
+	struct ip_vs_app *app;
 	int diff;
 	struct iphdr *iph;
 	struct tcphdr *th;
@@ -356,10 +439,10 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	 *	check if application module is bound to
 	 *	this ip_vs_conn.
 	 */
-	if ((vapp = cp->app) == NULL)
+	if ((app = cp->app) == NULL)
 		return 0;
 
-	iph = (*skb_p)->nh.iph;
+	iph = skb->nh.iph;
 	th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
 
 	/*
@@ -380,10 +463,10 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	/*
 	 *	Call private input hook function
 	 */
-	if (vapp->pkt_in == NULL)
+	if (app->pkt_in == NULL)
 		return 0;
 
-	diff = vapp->pkt_in(vapp, cp, skb_p);
+	diff = app->pkt_in(app, cp, skb);
 
 	/*
 	 *	Update ip_vs seq stuff if len has changed.
@@ -399,15 +482,14 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 /*
  *	/proc/net/ip_vs_app entry function
  */
-static int ip_vs_app_getinfo(char *buffer, char **start, off_t offset,
-			     int length)
+static int
+ip_vs_app_getinfo(char *buffer, char **start, off_t offset, int length)
 {
 	off_t pos=0;
 	int len=0;
 	char temp[64];
-	int idx;
-	struct ip_vs_app *vapp;
-	struct list_head *e;
+	struct ip_vs_app *app, *inc;
+	struct list_head *e, *i;
 
 	pos = 64;
 	if (pos > offset) {
@@ -415,26 +497,27 @@ static int ip_vs_app_getinfo(char *buffer, char **start, off_t offset,
 			       "prot port    usecnt name");
 	}
 
-	read_lock_bh(&__ip_vs_app_lock);
-	for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++) {
-		list_for_each (e, &ip_vs_app_base[idx]) {
-			vapp = list_entry(e, struct ip_vs_app, n_list);
+	down(&__ip_vs_app_mutex);
+	list_for_each (e, &ip_vs_app_list) {
+		app = list_entry(e, struct ip_vs_app, a_list);
+		list_for_each (i, &app->incs_list) {
+			inc = list_entry(i, struct ip_vs_app, a_list);
 
 			pos += 64;
 			if (pos <= offset)
 				continue;
 			sprintf(temp, "%-3s  %-7u %-6d %-17s",
-				ip_vs_proto_name(IP_VS_APP_PROTO(vapp->type)),
-				IP_VS_APP_PORT(vapp->type),
-				vapp->module?GET_USE_COUNT(vapp->module):0,
-				vapp->name);
+				ip_vs_proto_name(inc->protocol),
+				ntohs(inc->port),
+				atomic_read(&inc->usecnt),
+				inc->name);
 			len += sprintf(buffer+len, "%-63s\n", temp);
 			if (pos >= offset+length)
 				goto done;
 		}
 	}
   done:
-	read_unlock_bh(&__ip_vs_app_lock);
+	up(&__ip_vs_app_mutex);
 
 	*start = buffer+len-(pos-offset);       /* Start of wanted data */
 	len = pos-offset;
@@ -492,17 +575,12 @@ int ip_vs_skb_replace(struct sk_buff *skb, int pri,
 
 int ip_vs_app_init(void)
 {
-	int idx;
-
-	for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++) {
-		INIT_LIST_HEAD(&ip_vs_app_base[idx]);
-	}
-
 	/* we will replace it with proc_net_ipvs_create() soon */
 	proc_net_create("ip_vs_app", 0, ip_vs_app_getinfo);
 	return 0;
 }
 
+
 void ip_vs_app_cleanup(void)
 {
 	proc_net_remove("ip_vs_app");
