@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.31.2.5 2003/07/29 14:37:12 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.32 2002/09/15 08:14:08 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -29,6 +29,7 @@
 #include <linux/types.h>
 #include <linux/kernel.h>
 #include <linux/errno.h>
+#include <linux/compiler.h>
 #include <linux/ip.h>
 #include <linux/tcp.h>
 #include <linux/icmp.h>
@@ -53,7 +54,7 @@ EXPORT_SYMBOL(ip_vs_proto_name);
 EXPORT_SYMBOL(ip_vs_conn_new);
 EXPORT_SYMBOL(ip_vs_conn_in_get);
 EXPORT_SYMBOL(ip_vs_conn_out_get);
-EXPORT_SYMBOL(ip_vs_conn_listen);
+EXPORT_SYMBOL(ip_vs_tcp_conn_listen);
 EXPORT_SYMBOL(ip_vs_conn_put);
 #ifdef CONFIG_IP_VS_DEBUG
 EXPORT_SYMBOL(ip_vs_get_debug_level);
@@ -83,6 +84,11 @@ const char *ip_vs_proto_name(unsigned proto)
 	}
 }
 
+void ip_vs_init_hash_table(struct list_head *table, int rows)
+{
+	while (--rows >= 0)
+		INIT_LIST_HEAD(&table[rows]);
+}
 
 static inline void
 ip_vs_in_stats(struct ip_vs_conn *cp, struct sk_buff *skb)
@@ -146,11 +152,24 @@ ip_vs_conn_stats(struct ip_vs_conn *cp, struct ip_vs_service *svc)
 	spin_unlock(&ip_vs_stats.lock);
 }
 
+
+static inline int
+ip_vs_set_state(struct ip_vs_conn *cp, int direction,
+		struct iphdr *iph, union ip_vs_tphdr h,
+		struct ip_vs_protocol *pp)
+{
+	if (unlikely(!pp->state_transition))
+		return 0;
+	return pp->state_transition(cp, direction, iph, h, pp);
+}
+
+
 /*
  *  IPVS persistent scheduling function
  *  It creates a connection entry according to its template if exists,
  *  or selects a server and creates a connection entry plus a template.
  *  Locking: we are svc user (svc->refcnt), so we hold all dests too
+ *  Protocols supported: TCP, UDP
  */
 static struct ip_vs_conn *
 ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
@@ -167,7 +186,7 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 	/* Mask saddr with the netmask to adjust template granularity */
 	snet = iph->saddr & svc->netmask;
 
-	IP_VS_DBG(6, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
+	IP_VS_DBG(6, "p-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
 		  "mnet %u.%u.%u.%u\n",
 		  NIPQUAD(iph->saddr), ntohs(portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(portp[1]),
@@ -202,7 +221,7 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 			 */
 			dest = svc->scheduler->schedule(svc, iph);
 			if (dest == NULL) {
-				IP_VS_DBG(1, "P-schedule: no dest found.\n");
+				IP_VS_DBG(1, "p-schedule: no dest found.\n");
 				return NULL;
 			}
 
@@ -259,7 +278,7 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 
 			dest = svc->scheduler->schedule(svc, iph);
 			if (dest == NULL) {
-				IP_VS_DBG(1, "P-schedule: no dest found.\n");
+				IP_VS_DBG(1, "p-schedule: no dest found.\n");
 				return NULL;
 			}
 
@@ -305,19 +324,13 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 		return NULL;
 	}
 
-	/*
-	 *    Increase the inactive connection counter
-	 *    because it is in Syn-Received
-	 *    state (inactive) when the connection is created.
-	 */
-	atomic_inc(&dest->inactconns);
-
 	/*
 	 *    Add its control
 	 */
 	ip_vs_control_add(cp, ct);
-
 	ip_vs_conn_put(ct);
+
+	ip_vs_conn_stats(cp, svc);
 	return cp;
 }
 
@@ -326,8 +339,9 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
  *  IPVS main scheduling function
  *  It selects a server according to the virtual service, and
  *  creates a connection entry.
+ *  Protocols supported: TCP, UDP
  */
-static struct ip_vs_conn *
+struct ip_vs_conn *
 ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 {
 	struct ip_vs_conn *cp = NULL;
@@ -370,20 +384,15 @@ ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 	if (cp == NULL)
 		return NULL;
 
-	/*
-	 *    Increase the inactive connection counter because it is in
-	 *    Syn-Received state (inactive) when the connection is created.
-	 */
-	atomic_inc(&dest->inactconns);
-
-	IP_VS_DBG(6, "Schedule fwd:%c s:%s c:%u.%u.%u.%u:%u v:%u.%u.%u.%u:%u "
+	IP_VS_DBG(6, "Schedule fwd:%c c:%u.%u.%u.%u:%u v:%u.%u.%u.%u:%u "
 		  "d:%u.%u.%u.%u:%u flg:%X cnt:%d\n",
-		  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
+		  ip_vs_fwd_tag(cp),
 		  NIPQUAD(cp->caddr), ntohs(cp->cport),
 		  NIPQUAD(cp->vaddr), ntohs(cp->vport),
 		  NIPQUAD(cp->daddr), ntohs(cp->dport),
 		  cp->flags, atomic_read(&cp->refcnt));
 
+	ip_vs_conn_stats(cp, svc);
 	return cp;
 }
 
@@ -393,17 +402,17 @@ ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
  *  Called by ip_vs_in, when the virtual service is available but
  *  no destination is available for a new connection.
  */
-static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
+int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb,
+	       struct ip_vs_protocol *pp, union ip_vs_tphdr h)
 {
 	struct iphdr *iph = skb->nh.iph;
-	__u16 *portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);
 
 	/* if it is fwmark-based service, the cache_bypass sysctl is up
 	   and the destination is RTN_UNICAST (and not local), then create
 	   a cache_bypass connection entry */
 	if (sysctl_ip_vs_cache_bypass && svc->fwmark
 	    && (inet_addr_type(iph->daddr) == RTN_UNICAST)) {
-		int ret;
+		int ret, cs;
 		struct ip_vs_conn *cp;
 
 		ip_vs_service_put(svc);
@@ -411,8 +420,8 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 		/* create a new connection entry */
 		IP_VS_DBG(6, "ip_vs_leave: create a cache_bypass entry\n");
 		cp = ip_vs_conn_new(iph->protocol,
-				    iph->saddr, portp[0],
-				    iph->daddr, portp[1],
+				    iph->saddr, h.portp[0],
+				    iph->daddr, h.portp[1],
 				    0, 0,
 				    IP_VS_CONN_F_BYPASS,
 				    NULL);
@@ -425,10 +434,10 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 		ip_vs_in_stats(cp, skb);
 
 		/* set state */
-		ip_vs_set_state(cp, VS_STATE_INPUT, iph, portp);
+		cs = ip_vs_set_state(cp, IP_VS_DIR_INPUT, iph, h, pp);
 
 		/* transmit the first SYN packet */
-		ret = cp->packet_xmit(skb, cp);
+		ret = cp->packet_xmit(skb, cp, pp);
 
 		atomic_inc(&cp->in_pkts);
 		ip_vs_conn_put(cp);
@@ -441,7 +450,7 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 	 * listed in the ipvs table), pass the packets, because it is
 	 * not ipvs job to decide to drop the packets.
 	 */
-	if ((svc->port == FTPPORT) && (portp[1] != FTPPORT)) {
+	if ((svc->port == FTPPORT) && (h.portp[1] != FTPPORT)) {
 		ip_vs_service_put(svc);
 		return NF_ACCEPT;
 	}
@@ -491,19 +500,22 @@ static unsigned int ip_vs_post_routing(unsigned int hooknum,
  *	Find any that might be relevant, check against existing connections,
  *	forward to the right destination host if relevant.
  *	Currently handles error types - unreachable, quench, ttl exceeded.
- *      (Only used in VS/NAT)
+ *	(Only used in VS/NAT)
  */
-static int ip_vs_out_icmp(struct sk_buff **skb_p)
+static int ip_vs_out_icmp(struct sk_buff **skb_p, int *related)
 {
 	struct sk_buff	*skb   = *skb_p;
 	struct iphdr	*iph;
 	struct icmphdr	*icmph;
 	struct iphdr	*ciph;	/* The ip header contained within the ICMP */
-	__u16		*pptr;	/* port numbers from TCP/UDP contained header */
 	unsigned short	ihl;
 	unsigned short	len;
-	unsigned short	clen, csize;
+	unsigned short	clen, cihl;
 	struct ip_vs_conn *cp;
+	struct ip_vs_protocol *pp;
+	union ip_vs_tphdr h;
+
+	*related = 1;
 
 	/* reassemble IP fragments, but will it happen in ICMP packets?? */
 	if (skb->nh.iph->frag_off & __constant_htons(IP_MF|IP_OFFSET)) {
@@ -539,36 +551,34 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	 */
 	if ((icmph->type != ICMP_DEST_UNREACH) &&
 	    (icmph->type != ICMP_SOURCE_QUENCH) &&
-	    (icmph->type != ICMP_TIME_EXCEEDED))
+	    (icmph->type != ICMP_TIME_EXCEEDED)) {
+		*related = 0;
 		return NF_ACCEPT;
+	}
 
 	/* Now find the contained IP header */
 	clen = len - sizeof(struct icmphdr);
 	if (clen < sizeof(struct iphdr))
 		return NF_DROP;
 	ciph = (struct iphdr *) (icmph + 1);
-	csize = ciph->ihl << 2;
-	if (clen < csize)
+	cihl = ciph->ihl << 2;
+	if (clen < cihl)
 		return NF_DROP;
 
-	/* We are only interested ICMPs generated from TCP or UDP packets */
-	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP)
+	pp = ip_vs_proto_get(ciph->protocol);
+	if (!pp)
 		return NF_ACCEPT;
 
-	/* Skip non-first embedded TCP/UDP fragments */
-	if (ciph->frag_off & __constant_htons(IP_OFFSET))
+	/* Is the embedded protocol header present? */
+	if (unlikely(ciph->frag_off & __constant_htons(IP_OFFSET) &&
+		     (pp->minhlen || pp->dont_defrag)))
 		return NF_ACCEPT;
 
 	/* We need at least TCP/UDP ports here */
-	if (clen < csize + sizeof(struct udphdr))
+	if (clen < cihl + pp->minhlen_icmp)
 		return NF_DROP;
 
-	/*
-	 * Find the ports involved - this packet was
-	 * incoming so the ports are right way round
-	 * (but reversed relative to outer IP header!)
-	 */
-	pptr = (__u16 *)&(((char *)ciph)[csize]);
+	h.raw = (char *) ciph + cihl;
 
 	/* Ensure the checksum is correct */
 	if (ip_compute_csum((unsigned char *) icmph, len)) {
@@ -578,14 +588,10 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 	}
 
-	IP_VS_DBG(11, "Handling outgoing ICMP for "
-		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
-		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
-		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
+	IP_VS_DBG_PKT(11, pp, ciph, "Handling outgoing ICMP for");
 
 	/* ciph content is actually <protocol, caddr, cport, daddr, dport> */
-	cp = ip_vs_conn_out_get(ciph->protocol, ciph->daddr, pptr[1],
-				ciph->saddr, pptr[0]);
+	cp = pp->conn_out_get(skb, pp, ciph, h, 1);
 	if (!cp)
 		return NF_ACCEPT;
 
@@ -604,7 +610,8 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	ip_send_check(ciph);
 
 	/* the TCP/UDP dest port - cannot redo check */
-	pptr[1] = cp->vport;
+	if (IPPROTO_TCP == ciph->protocol || IPPROTO_UDP == ciph->protocol)
+		h.portp[1] = cp->vport;
 
 	/* And finally the ICMP checksum */
 	icmph->checksum = 0;
@@ -613,12 +620,9 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 
 	/* do the statistics and put it back */
 	ip_vs_out_stats(cp, skb);
-	ip_vs_conn_put(cp);
+	__ip_vs_conn_put(cp);
 
-	IP_VS_DBG(11, "Forwarding correct outgoing ICMP to "
-		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
-		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
-		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
+	IP_VS_DBG_PKT(11, pp, ciph, "Forwarding correct outgoing ICMP");
 
 	skb->nfcache |= NFC_IPVS_PROPERTY;
 
@@ -631,18 +635,17 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
  *	Check if outgoing packet belongs to the established ip_vs_conn,
  *      rewrite addresses of the packet and send it on its way...
  */
-static unsigned int ip_vs_out(unsigned int hooknum,
-			      struct sk_buff **skb_p,
-			      const struct net_device *in,
-			      const struct net_device *out,
-			      int (*okfn)(struct sk_buff *))
+static unsigned int
+ip_vs_out(unsigned int hooknum, struct sk_buff **skb_p,
+	  const struct net_device *in, const struct net_device *out,
+	  int (*okfn)(struct sk_buff *))
 {
 	struct sk_buff  *skb = *skb_p;
 	struct iphdr	*iph;
+	struct ip_vs_protocol *pp;
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
-	int size;
-	int ihl;
+	int size, ihl, firstfrag;
 
 	EnterFunction(11);
 
@@ -650,15 +653,20 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		return NF_ACCEPT;
 
 	iph = skb->nh.iph;
-	if (iph->protocol == IPPROTO_ICMP)
-		return ip_vs_out_icmp(skb_p);
+	if (unlikely(iph->protocol == IPPROTO_ICMP)) {
+		int related, verdict = ip_vs_out_icmp(skb_p, &related);
 
-	/* let it go if other IP protocols */
-	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
+		if (related)
+			return verdict;
+	}
+
+	pp = ip_vs_proto_get(iph->protocol);
+	if (unlikely(!pp))
 		return NF_ACCEPT;
 
 	/* reassemble IP fragments */
-	if (iph->frag_off & __constant_htons(IP_MF|IP_OFFSET)) {
+	if (unlikely(iph->frag_off & __constant_htons(IP_MF|IP_OFFSET) &&
+		     !pp->dont_defrag)) {
 		skb = ip_defrag(skb);
 		if (!skb)
 			return NF_STOLEN;
@@ -666,22 +674,29 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		*skb_p = skb;
 	}
 
-	/* make sure that protocol header available in skb data area,
+	/* make sure that protocol header is available in skb data area,
 	   note that skb data area may be reallocated. */
 	ihl = iph->ihl << 2;
-	if (ip_vs_header_check(skb, iph->protocol, ihl) == -1)
+	firstfrag = !(iph->frag_off & __constant_htons(IP_OFFSET));
+	/*
+	 * WARNING: we can work with !firstfrag packets, make sure
+	 * each protocol handler checks for firstfrag
+	 */
+	if (firstfrag &&
+	    !pskb_may_pull(skb, ihl+pp->minhlen))
 		return NF_DROP;
-
 	iph = skb->nh.iph;
 	h.raw = (char*) iph + ihl;
 
 	/*
-	 *	Check if the packet belongs to an old entry
+	 * Check if the packet belongs to an existing entry
 	 */
-	cp = ip_vs_conn_out_get(iph->protocol, iph->saddr, h.portp[0],
-				iph->daddr, h.portp[1]);
-	if (!cp) {
+	cp = pp->conn_out_get(skb, pp, iph, h, 0);
+
+	if (unlikely(!cp)) {
 		if (sysctl_ip_vs_nat_icmp_send &&
+		    (pp->protocol == IPPROTO_TCP ||
+		     pp->protocol == IPPROTO_UDP) &&
 		    ip_vs_lookup_real_service(iph->protocol,
 					      iph->saddr, h.portp[0])) {
 			/*
@@ -695,12 +710,9 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 				return NF_STOLEN;
 			}
 		}
-		IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d "
-			  "continue traversal as normal.\n",
-			  ip_vs_proto_name(iph->protocol),
-			  NIPQUAD(iph->daddr),
-			  ntohs(h.portp[1]));
-		if (skb_is_nonlinear(skb))
+		IP_VS_DBG_PKT(12, pp, iph,
+			"packet continues traversal as normal");
+		if (!pp->dont_defrag)
 			ip_send_check(iph);
 		return NF_ACCEPT;
 	}
@@ -713,7 +725,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	 * do final checksum checking.
 	 */
 
-	if (cp->app && skb_is_nonlinear(skb)) {
+	if (unlikely(cp->app && !pp->slave && skb_is_nonlinear(skb))) {
 		if (skb_linearize(skb, GFP_ATOMIC) != 0) {
 			ip_vs_conn_put(cp);
 			return NF_DROP;
@@ -723,91 +735,31 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	}
 
 	size = skb->len - ihl;
-	IP_VS_DBG(11, "O-pkt: %s size=%d\n",
-		  ip_vs_proto_name(iph->protocol), size);
+	IP_VS_DBG(11, "O-pkt: %s size=%d\n", pp->name, size);
 
 	/* do TCP/UDP checksum checking if it has application helper */
-	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		switch (skb->ip_summed) {
-		case CHECKSUM_NONE:
-			skb->csum = csum_partial(h.raw, size, 0);
-		case CHECKSUM_HW:
-			if (csum_tcpudp_magic(iph->saddr, iph->daddr, size,
-					      iph->protocol, skb->csum)) {
-				ip_vs_conn_put(cp);
-				IP_VS_DBG_RL("Outgoing failed %s checksum "
-					     "from %d.%d.%d.%d (size=%d)!\n",
-					     ip_vs_proto_name(iph->protocol),
-					     NIPQUAD(iph->saddr),
-					     size);
-				return NF_DROP;
-			}
-			break;
-		default:
-			/* CHECKSUM_UNNECESSARY */
-			break;
+	if (unlikely(cp->app && pp->csum_check && !pp->slave)) {
+		if (!pp->csum_check(skb, pp, iph, h, size)) {
+			ip_vs_conn_put(cp);
+			return NF_DROP;
 		}
 	}
 
-	IP_VS_DBG(11, "Outgoing %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
-		  ip_vs_proto_name(iph->protocol),
-		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
-		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
+	IP_VS_DBG_PKT(11, pp, iph, "Outgoing packet");
 
 	/* mangle the packet */
 	iph->saddr = cp->vaddr;
-	h.portp[0] = cp->vport;
-
-	/*
-	 *	Call application helper if needed
-	 */
-	if (ip_vs_app_pkt_out(cp, skb) != 0) {
-		/* skb data has probably changed, update pointers */
+	if (pp->snat_handler) {
+		pp->snat_handler(skb, pp, cp, iph, h, size);
 		iph = skb->nh.iph;
-		h.raw = (char*)iph + ihl;
-		size = skb->len - ihl;
-	}
-
-	/*
-	 *	Adjust TCP/UDP checksums
-	 */
-	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		/* Only port and addr are changed, do fast csum update */
-		ip_vs_fast_check_update(&h, cp->daddr, cp->vaddr,
-					cp->dport, cp->vport, iph->protocol);
-		if (skb->ip_summed == CHECKSUM_HW)
-			skb->ip_summed = CHECKSUM_NONE;
-	} else {
-		/* full checksum calculation */
-		switch (iph->protocol) {
-		case IPPROTO_TCP:
-			h.th->check = 0;
-			skb->csum = csum_partial(h.raw, size, 0);
-			h.th->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
-							size, iph->protocol,
-							skb->csum);
-			IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
-				  ip_vs_proto_name(iph->protocol), h.th->check,
-				  (char*)&(h.th->check) - (char*)h.raw);
-			break;
-		case IPPROTO_UDP:
-			h.uh->check = 0;
-			skb->csum = csum_partial(h.raw, size, 0);
-			h.uh->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
-							size, iph->protocol,
-							skb->csum);
-			if (h.uh->check == 0)
-				h.uh->check = 0xFFFF;
-			IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
-				  ip_vs_proto_name(iph->protocol), h.uh->check,
-				  (char*)&(h.uh->check) - (char*)h.raw);
-			break;
-		}
+		h.raw = (char*) iph + ihl;
 	}
 	ip_send_check(iph);
 
+	IP_VS_DBG_PKT(10, pp, iph, "After SNAT");
+
 	ip_vs_out_stats(cp, skb);
-	ip_vs_set_state(cp, VS_STATE_OUTPUT, iph, h.portp);
+	ip_vs_set_state(cp, IP_VS_DIR_OUTPUT, iph, h, pp);
 	ip_vs_conn_put(cp);
 
 	skb->nfcache |= NFC_IPVS_PROPERTY;
@@ -844,33 +796,33 @@ unsigned int check_for_ip_vs_out(struct sk_buff **skb_p,
 
 
 /*
- *	Handle ICMP messages in the outside-to-inside direction (incoming)
- *	and sometimes in outgoing direction from ip_vs_forward_icmp.
+ *	Handle ICMP messages in the outside-to-inside direction (incoming).
  *	Find any that might be relevant, check against existing connections,
  *	forward to the right destination host if relevant.
- *	Currently handles error types - unreachable, quench, ttl exceeded.
+ *	Currently handles error types - unreachable, quench, ttl exceeded
  */
-static int ip_vs_in_icmp(struct sk_buff **skb_p)
+static int ip_vs_in_icmp(struct sk_buff **skb_p, int *related)
 {
-	struct sk_buff	*skb   = *skb_p;
+	struct sk_buff	*skb = *skb_p;
 	struct iphdr    *iph;
 	struct icmphdr  *icmph;
 	struct iphdr    *ciph;	/* The ip header contained within the ICMP */
-	__u16	        *pptr;	/* port numbers from TCP/UDP contained header */
-	unsigned short   len;
-	unsigned short	clen, csize;
+	unsigned short	len;
+	unsigned short	clen, cihl;
 	struct ip_vs_conn *cp;
-	struct rtable *rt;			/* Route to the other host */
-	int    mtu;
+	struct ip_vs_protocol *pp;
+	union ip_vs_tphdr h;
+	int	rc;
 
+	*related = 1;
 	if (skb_is_nonlinear(skb)) {
 		if (skb_linearize(skb, GFP_ATOMIC) != 0)
 			return NF_DROP;
+		ip_send_check(skb->nh.iph);
 	}
 
 	iph = skb->nh.iph;
-	ip_send_check(iph);
-	icmph = (struct icmphdr *)((char *)iph + (iph->ihl << 2));
+	icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
 	len = ntohs(iph->tot_len) - (iph->ihl<<2);
 	if (len < sizeof(struct icmphdr))
 		return NF_DROP;
@@ -881,8 +833,10 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 
 	if ((icmph->type != ICMP_DEST_UNREACH) &&
 	    (icmph->type != ICMP_SOURCE_QUENCH) &&
-	    (icmph->type != ICMP_TIME_EXCEEDED))
+	    (icmph->type != ICMP_TIME_EXCEEDED)) {
+		*related = 0;
 		return NF_ACCEPT;
+	}
 
 	/*
 	 * If we get here we have an ICMP error of one of the above 3 types
@@ -892,20 +846,21 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	if (clen < sizeof(struct iphdr))
 		return NF_DROP;
 	ciph = (struct iphdr *) (icmph + 1);
-	csize = ciph->ihl << 2;
-	if (clen < csize)
+	cihl = ciph->ihl << 2;
+	if (clen < cihl)
 		return NF_DROP;
 
-	/* We are only interested ICMPs generated from TCP or UDP packets */
-	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP)
+	pp = ip_vs_proto_get(ciph->protocol);
+	if (!pp)
 		return NF_ACCEPT;
 
-	/* Skip non-first embedded TCP/UDP fragments */
-	if (ciph->frag_off & __constant_htons(IP_OFFSET))
+	/* Is the embedded protocol header present? */
+	if (unlikely(ciph->frag_off & __constant_htons(IP_OFFSET) &&
+		     (pp->minhlen || pp->dont_defrag)))
 		return NF_ACCEPT;
 
 	/* We need at least TCP/UDP ports here */
-	if (clen < csize + sizeof(struct udphdr))
+	if (clen < cihl + pp->minhlen_icmp)
 		return NF_DROP;
 
 	/* Ensure the checksum is correct */
@@ -916,102 +871,20 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 	}
 
-	pptr = (__u16 *)&(((char *)ciph)[csize]);
+	h.raw = (char *) ciph + cihl;
 
-	IP_VS_DBG(11, "Handling incoming ICMP for "
-		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
-		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
-		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
+	IP_VS_DBG_PKT(11, pp, ciph, "Handling incoming ICMP for");
 
 	/* This is pretty much what ip_vs_conn_in_get() does,
 	   except parameters are in the reverse order */
-	cp = ip_vs_conn_in_get(ciph->protocol,
-			       ciph->daddr, pptr[1],
-			       ciph->saddr, pptr[0]);
+	cp = pp->conn_in_get(skb, pp, ciph, h, 1);
 	if (cp == NULL)
 		return NF_ACCEPT;
 
 	ip_vs_in_stats(cp, skb);
-
-	/* The ICMP packet for VS/TUN, VS/DR and LOCALNODE will be
-	   forwarded directly here, because there is no need to
-	   translate address/port back */
-	if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ) {
-		int ret;
-		if (cp->packet_xmit)
-			ret = cp->packet_xmit(skb, cp);
-		else
-			ret = NF_ACCEPT;
-		atomic_inc(&cp->in_pkts);
-		ip_vs_conn_put(cp);
-		return ret;
-	}
-
-	/*
-	 * mangle and send the packet here
-	 */
-	if (!(rt = __ip_vs_get_out_rt(cp, RT_TOS(iph->tos))))
-		goto tx_error_icmp;
-
-	/* MTU checking */
-	mtu = rt->u.dst.pmtu;
-	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
-		ip_rt_put(rt);
-		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
-		IP_VS_DBG_RL("ip_vs_in_icmp(): frag needed\n");
-		goto tx_error;
-	}
-
-	/* drop old route */
-	dst_release(skb->dst);
-	skb->dst = &rt->u.dst;
-
-	/* copy-on-write the packet before mangling it */
-	if (ip_vs_skb_cow(skb, rt->u.dst.dev->hard_header_len,
-			  &iph, (unsigned char**)&icmph)) {
-		ip_vs_conn_put(cp);
-		return NF_DROP;
-	}
-	ciph = (struct iphdr *) (icmph + 1);
-	pptr = (__u16 *)&(((char *)ciph)[csize]);
-
-	/* The ICMP packet for VS/NAT must be written to correct addresses
-	   before being forwarded to the right server */
-
-	/* First change the dest IP address, and recalc checksum */
-	iph->daddr = cp->daddr;
-	ip_send_check(iph);
-
-	/* Now change the *source* address in the contained IP */
-	ciph->saddr = cp->daddr;
-	ip_send_check(ciph);
-
-	/* the TCP/UDP source port - cannot redo check */
-	pptr[0] = cp->dport;
-
-	/* And finally the ICMP checksum */
-	icmph->checksum = 0;
-	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
-	skb->ip_summed = CHECKSUM_UNNECESSARY;
-
-	IP_VS_DBG(11, "Forwarding incoming ICMP to "
-		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
-		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
-		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
-
-#ifdef CONFIG_NETFILTER_DEBUG
-	skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
-#endif /* CONFIG_NETFILTER_DEBUG */
-	ip_send(skb);
-	ip_vs_conn_put(cp);
-	return NF_STOLEN;
-
-  tx_error_icmp:
-	dst_link_failure(skb);
-  tx_error:
-	dev_kfree_skb(skb);
-	ip_vs_conn_put(cp);
-	return NF_STOLEN;
+	rc = ip_vs_icmp_xmit(skb, cp, pp);
+	__ip_vs_conn_put(cp);
+	return rc;
 }
 
 
@@ -1027,17 +900,17 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 {
 	struct sk_buff	*skb = *skb_p;
 	struct iphdr	*iph = skb->nh.iph;
+	struct ip_vs_protocol *pp = ip_vs_proto_get(iph->protocol);
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
-	struct ip_vs_service *svc;
-	int ihl;
-	int ret;
+	int ihl, ret, restart;
+	int firstfrag;
 
 	/*
-	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
+	 *	Big tappo: only PACKET_HOST (neither loopback nor mcasts)
 	 *	... don't know why 1st test DOES NOT include 2nd (?)
 	 */
-	if (skb->pkt_type != PACKET_HOST || skb->dev == &loopback_dev) {
+	if (unlikely(skb->pkt_type != PACKET_HOST || skb->dev == &loopback_dev)) {
 		IP_VS_DBG(12, "packet type=%d proto=%d daddr=%d.%d.%d.%d ignored\n",
 			  skb->pkt_type,
 			  iph->protocol,
@@ -1045,17 +918,32 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		return NF_ACCEPT;
 	}
 
-	if (iph->protocol == IPPROTO_ICMP)
-		return ip_vs_in_icmp(skb_p);
+	if (unlikely(iph->protocol == IPPROTO_ICMP)) {
+		int related, verdict = ip_vs_in_icmp(skb_p, &related);
 
-	/* let it go if other IP protocols */
-	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
+		if (related)
+			return verdict;
+	}
+
+	/* Protocol supported? */
+	if (unlikely(!pp))
 		return NF_ACCEPT;
 
-	/* make sure that protocol header available in skb data area,
+	/* make sure that protocol header is available in skb data area,
 	   note that skb data area may be reallocated. */
 	ihl = iph->ihl << 2;
-	if (ip_vs_header_check(skb, iph->protocol, ihl) == -1)
+#if 0
+	/* Enable this when not in LOCAL_IN */
+	firstfrag = !(iph->frag_off & __constant_htons(IP_OFFSET));
+	/*
+	 * WARNING: we can work with !firstfrag packets, make sure
+	 * each protocol handler checks for firstfrag
+	 */
+#else
+	firstfrag = 1;
+#endif
+	if (firstfrag &&
+	    !pskb_may_pull(skb, ihl+pp->minhlen))
 		return NF_DROP;
 	iph = skb->nh.iph;
 	h.raw = (char*) iph + ihl;
@@ -1063,47 +951,24 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	/*
 	 * Check if the packet belongs to an existing connection entry
 	 */
-	cp = ip_vs_conn_in_get(iph->protocol, iph->saddr, h.portp[0],
-			       iph->daddr, h.portp[1]);
+	cp = pp->conn_in_get(skb, pp, iph, h, 0);
 
-	if (!cp &&
-	    (h.th->syn || (iph->protocol!=IPPROTO_TCP)) &&
-	    (svc = ip_vs_service_get(skb->nfmark, iph->protocol,
-				     iph->daddr, h.portp[1]))) {
-		if (ip_vs_todrop()) {
-			/*
-			 * It seems that we are very loaded.
-			 * We have to drop this packet :(
-			 */
-			ip_vs_service_put(svc);
-			return NF_DROP;
+	if (unlikely(!cp)) {
+		int v;
+
+		if (!pp->conn_schedule(skb, pp, iph, h, &v, &cp)) {
+			return v;
 		}
-
-		/*
-		 * Let the virtual server select a real server for the
-		 * incoming connection, and create a connection entry.
-		 */
-		cp = ip_vs_schedule(svc, iph);
-		if (!cp)
-			return ip_vs_leave(svc, skb);
-		ip_vs_conn_stats(cp, svc);
-		ip_vs_service_put(svc);
 	}
 
-	if (!cp) {
+	if (unlikely(!cp)) {
 		/* sorry, all this trouble for a no-hit :) */
-		IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d continue "
-			  "traversal as normal.\n",
-			  ip_vs_proto_name(iph->protocol),
-			  NIPQUAD(iph->daddr),
-			  ntohs(h.portp[1]));
+		IP_VS_DBG_PKT(12, pp, iph,
+			"packet continues traversal as normal");
 		return NF_ACCEPT;
 	}
 
-	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
-		  ip_vs_proto_name(iph->protocol),
-		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
-		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
+	IP_VS_DBG_PKT(11, pp, iph, "Incoming packet");
 
 	/* Check the server status */
 	if (cp->dest && !(cp->dest->flags & IP_VS_DEST_F_AVAILABLE)) {
@@ -1121,9 +986,9 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	}
 
 	ip_vs_in_stats(cp, skb);
-	ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp);
+	restart = ip_vs_set_state(cp, IP_VS_DIR_INPUT, iph, h, pp);
 	if (cp->packet_xmit)
-		ret = cp->packet_xmit(skb, cp);
+		ret = cp->packet_xmit(skb, cp, pp);
 	else {
 		IP_VS_DBG_RL("warning: packet_xmit is null");
 		ret = NF_ACCEPT;
@@ -1134,7 +999,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	atomic_inc(&cp->in_pkts);
 	if (ip_vs_sync_state == IP_VS_STATE_MASTER &&
 	    (cp->protocol != IPPROTO_TCP ||
-	     cp->state == IP_VS_S_ESTABLISHED) &&
+	     cp->state == IP_VS_TCP_S_ESTABLISHED) &&
 	    (atomic_read(&cp->in_pkts) % 50 == sysctl_ip_vs_sync_threshold))
 		ip_vs_sync_conn(cp);
 
@@ -1160,6 +1025,7 @@ static unsigned int ip_vs_forward_icmp(unsigned int hooknum,
 {
 	struct sk_buff	*skb = *skb_p;
 	struct iphdr	*iph = skb->nh.iph;
+	int r;
 
 	if (iph->protocol != IPPROTO_ICMP)
 		return NF_ACCEPT;
@@ -1171,7 +1037,7 @@ static unsigned int ip_vs_forward_icmp(unsigned int hooknum,
 		*skb_p = skb;
 	}
 
-	return ip_vs_in_icmp(skb_p);
+	return ip_vs_in_icmp(skb_p, &r);
 }
 
 
@@ -1216,23 +1082,27 @@ static int __init ip_vs_init(void)
 		goto cleanup_nothing;
 	}
 
-	ret = ip_vs_conn_init();
-	if (ret < 0) {
-		IP_VS_ERR("can't setup connection table.\n");
-		goto cleanup_control;
-	}
+	ip_vs_sltimer_init();
+	ip_vs_protocol_init();
 
 	ret = ip_vs_app_init();
 	if (ret < 0) {
 		IP_VS_ERR("can't setup application helper.\n");
-		goto cleanup_conn;
+		goto cleanup_protocol;
+	}
+
+	ret = ip_vs_conn_init();
+	if (ret < 0) {
+		IP_VS_ERR("can't setup connection table.\n");
+		goto cleanup_app;
 	}
 
 	ret = nf_register_hook(&ip_vs_in_ops);
 	if (ret < 0) {
 		IP_VS_ERR("can't register in hook.\n");
-		goto cleanup_app;
+		goto cleanup_conn;
 	}
+
 	ret = nf_register_hook(&ip_vs_out_ops);
 	if (ret < 0) {
 		IP_VS_ERR("can't register out hook.\n");
@@ -1258,11 +1128,13 @@ static int __init ip_vs_init(void)
 	nf_unregister_hook(&ip_vs_out_ops);
   cleanup_inops:
 	nf_unregister_hook(&ip_vs_in_ops);
-  cleanup_app:
-	ip_vs_app_cleanup();
   cleanup_conn:
 	ip_vs_conn_cleanup();
-  cleanup_control:
+  cleanup_app:
+	ip_vs_app_cleanup();
+  cleanup_protocol:
+	ip_vs_protocol_cleanup();
+	ip_vs_sltimer_cleanup();
 	ip_vs_control_cleanup();
   cleanup_nothing:
 	return ret;
@@ -1274,8 +1146,10 @@ static void __exit ip_vs_cleanup(void)
 	nf_unregister_hook(&ip_vs_post_routing_ops);
 	nf_unregister_hook(&ip_vs_out_ops);
 	nf_unregister_hook(&ip_vs_in_ops);
-	ip_vs_app_cleanup();
 	ip_vs_conn_cleanup();
+	ip_vs_app_cleanup();
+	ip_vs_protocol_cleanup();
+	ip_vs_sltimer_cleanup();
 	ip_vs_control_cleanup();
 	IP_VS_INFO("ipvs unloaded.\n");
 }
