@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sync.c,v 1.8 2002/08/17 14:06:02 wensong Exp $
+ * Version:     $Id: ip_vs_sync.c,v 1.9 2002/09/15 08:14:08 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -24,6 +24,7 @@
 #include <linux/sched.h>
 #include <linux/wait.h>
 #include <linux/unistd.h>
+#include <linux/completion.h>
 
 #include <linux/skbuff.h>
 #include <linux/in.h>
@@ -325,9 +326,11 @@ static struct sockaddr_in mcast_addr;
  */
 static void set_mcast_loop(struct sock *sk, u_char loop)
 {
+	struct inet_opt *inet = inet_sk(sk);
+
 	/* setsockopt(sock, SOL_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)); */
 	lock_sock(sk);
-	sk->protinfo.af_inet.mc_loop = loop ? 1 : 0;
+	inet->mc_loop = loop ? 1 : 0;
 	release_sock(sk);
 }
 
@@ -336,9 +339,11 @@ static void set_mcast_loop(struct sock *sk, u_char loop)
  */
 static void set_mcast_ttl(struct sock *sk, u_char ttl)
 {
+	struct inet_opt *inet = inet_sk(sk);
+
 	/* setsockopt(sock, SOL_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)); */
 	lock_sock(sk);
-	sk->protinfo.af_inet.mc_ttl = ttl;
+	inet->mc_ttl = ttl;
 	release_sock(sk);
 }
 
@@ -348,6 +353,7 @@ static void set_mcast_ttl(struct sock *sk, u_char ttl)
 static int set_mcast_if(struct sock *sk, char *ifname)
 {
 	struct net_device *dev;
+	struct inet_opt *inet = inet_sk(sk);
 
 	if ((dev = __dev_get_by_name(ifname)) == NULL)
 		return -ENODEV;
@@ -356,8 +362,8 @@ static int set_mcast_if(struct sock *sk, char *ifname)
 		return -EINVAL;
 
 	lock_sock(sk);
-	sk->protinfo.af_inet.mc_index = dev->ifindex;
-	/*  sk->protinfo.af_inet.mc_addr  = 0; */
+	inet->mc_index = dev->ifindex;
+	/*  inet->mc_addr  = 0; */
 	release_sock(sk);
 
 	return 0;
@@ -686,7 +692,7 @@ static int sync_thread(void *startup)
 
 	spin_lock_irq(&current->sigmask_lock);
 	siginitsetinv(&current->blocked, 0);
-	recalc_sigpending(current);
+	recalc_sigpending();
 	spin_unlock_irq(&current->sigmask_lock);
 
 	/* set up multicast address */
